"""
Tucan module to get the imports of a fortran file
"""
from loguru import logger
from typing import List
from tucan.string_utils import tokenize
from tucan.unformat_common import Statements
from tucan.imports_c import by_cpp_include


def _by_use(tokens):
    """Resolve a Fortran use import"""
    # namespaces

    re_tokens = [
        token for token in tokens if token not in [",", ":", "::"]
    ]

    out = {}
    if "intrinsic" in re_tokens or "INTRINSIC" in re_tokens:
        for i in range(2, len(re_tokens), 1):
            #out[re_tokens[i]] = f"__intrinsic__.{re_tokens[i]}"
            out[re_tokens[i]] = f"__ftn__@__intrinsic__.{re_tokens[i]}"
        return out

    ref = re_tokens[1]
    if "only" in re_tokens or "ONLY" in re_tokens:
        for i in range(3, len(re_tokens), 1):
            #out[re_tokens[i]] = f"{ref}.{re_tokens[i]}"
            out[re_tokens[i]] = f"__ftn__@{ref}.{re_tokens[i]}"
            
    else:
        #out[f"_{ref}"] = ref
        out[f"_{ref}"] = f"__ftn__@{ref}"

    return out


def _by_include(tokens):
    """Resolve a Fortran include  import"""
    # namespaces
    ref = tokens[1].strip("'").strip('"')
    #return {f"_{ref}": ref}
    return {f"_{ref}": ref+"@*"}


def get_imports_ftn(lines: List[str],verbose=False) -> dict:
    """Read the imports information from python lines"""
    out = {}
    for i,line in enumerate(lines):
        if line.lstrip().replace(" ","").startswith(("use", "#include", "include","USE", "#INCLUDE", "INCLUDE")):
            tokens = tokenize(line)
            try:
                if tokens[0] in ["use", "USE"]:
                    out.update(_by_use(tokens))
                elif tokens[0] in ["include", "INCLUDE"]:
                    out.update(_by_include(tokens))
                elif tokens[1] in ["include", "INCLUDE"]:
                    out.update(by_cpp_include(tokens))
            except IndexError:
                pass

    return out



from loguru import logger
from typing import List

from tucan.unformat_common import (
    Statements,
    new_stmts,
    remove_strings,
    clean_blanks,
    clean_inline_comments,
    split_multi_statement_lines,
)
from tucan.string_utils import eat_spaces, get_indent

COMB_CHARS = {"'''": "\xaa", '"""': "\xbb"}


def remove_combined_chars(code: List[str]) -> List[str]:
    """replace combined chars with single char alternatives"""
    new_code = []
    for line in code:
        rline = line
        for cchar, rchar in COMB_CHARS.items():
            rline = rline.replace(cchar, rchar)
        new_code.append(rline)
    return new_code


def getback_combined_chars(code: List[str]) -> List[str]:
    """replace combined chars with single char alternatives, back"""
    new_code = []
    for line in code:
        rline = line
        for cchar, rchar in COMB_CHARS.items():
            rline = rline.replace(rchar, cchar)
        new_code.append(rline)
    return new_code


def align_continuation_lines(stmts: Statements) -> Statements:
    """Merge lines using \\ symbol at continuation line"""
    new_stmt = []
    new_lines = []
    last_line = ""
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if last_line.endswith("\\"):
            new_stmt[-1] = last_line[:-1] + " " + line.strip()
            new_lines[-1][1] = lend
        else:
            new_stmt.append(line)
            new_lines.append([lstart, lend])

        last_line = new_stmt[-1]
    return Statements(new_stmt, new_lines)


def align_multiline_strings(stmts: Statements, markup: str) -> Statements:
    new_stmt = []
    new_lines = []
    buffer = ""

    in_multiline = False
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if in_multiline:
            line = " " + line.lstrip()

        if not in_multiline:
            last_start = lstart

        for char in line:
            if char == markup:
                if not in_multiline:
                    in_multiline = True
                else:
                    in_multiline = False
            buffer += char
        # end of line reached...

        if not in_multiline:
            new_stmt.append(buffer)
            new_lines.append([last_start, lend])
            buffer = ""

    return Statements(new_stmt, new_lines)


def align_context_block(stmts: Statements) -> Statements:
    """
    Join statements on [({})]

    Args:
        stmts (stmts): List of unformatted code statements along with line number ranges.
    Returns:
        Statements: List of unformatted code statements along with line number ranges. Everything in
            a parenthesis is now a one liner.
    """
    new_stmt = []
    new_lines = []
    inside_block_context = []
    buffer = ""
    for line, (lstart, lend) in zip(stmts.stmt, stmts.lines):
        if not inside_block_context:
            this_start = lstart
            indent = get_indent(line)

        for char in line.lstrip():
            buffer += char
            if char in [
                "(",
                "[",
                "{",
            ]:
                inside_block_context.append(char)
                continue
            if char in [")", "]", "}"]:
                try:
                    inside_block_context.pop(-1)
                except IndexError:
                    logger.warning(
                        f"line {lstart} , unmatched context closing symbol -{char}-"
                    )
                continue

        # EOL reached !

        if not inside_block_context:
            new_stmt.append(indent + buffer)
            new_lines.append([this_start, lend])
            buffer = ""

    return Statements(new_stmt, new_lines)


def replace_keywords(code: List[str]) -> List[str]:
    """replace some uneasy keywords for parsing"""
    new_stmt = []

    for line in code:
        new_line = line.replace("super()", "self_super")
        new_stmt.append(new_line)
    return new_stmt


def split_fstrings(code: List[str], fstring_char: str, other_char: str) -> List[str]:
    """Split f-string into multiple strings before the remove_strings()

    for example:
     'np f" lorem {ipsum} sic " hamet '
     becomes  (split_fstrings)
     'no f" lorem "{ipsum}" sic " hamet '
     then (remove_strings)
     'no f"___"{ipsum}"___" hamet '

    """
    new_stmt = []

    for line in code:
        inside_fstring_context = False
        # inside_escape_context=False
        buffer = ""
        last_char = " "
        for char in line:
            if char == fstring_char:
                if last_char == "f":
                    inside_fstring_context = True
                else:
                    inside_fstring_context = False

            if inside_fstring_context and char == other_char:
                char = "_"

            buffer += char

            if inside_fstring_context:
                if char == "{":
                    buffer = buffer[:-1] + fstring_char + "{"
                    # inside_escape_context=True
                elif char == "}":
                    # buffer+=char
                    buffer += fstring_char  # insert one quotation after }
                    # inside_escape_context=True

            last_char = char

        new_stmt.append(buffer)
    return new_stmt


def enforce_4char_indents(code: List[str]) -> List[str]:
    """for s***o***b***  using somthg else than 4char indents"""
    smallest_indent = 1000
    for line in code:
        recess = len(get_indent(line))
        if recess > 0 and recess < smallest_indent:
            smallest_indent = recess
    if smallest_indent in [4, 1000]:
        return code

    new_code = []
    for line in code:
        recess = int(len(get_indent(line)) / smallest_indent)
        new_code.append("    " * recess + line.lstrip())
    return new_code


def unformat_py(code: List[str]) -> Statements:
    """
    Unformat Python code by stripping comments and breaking multiline statements.

    Args:
        code (List[str]): List of Python code lines.

    Returns:
        Statements: List of unformatted code statements along with line number ranges.
    """
    stmts = new_stmts(code)
    stmts.stmt = eat_spaces(stmts.stmt)
    stmts.stmt = remove_combined_chars(stmts.stmt)
    stmts = align_continuation_lines(stmts)
    stmts = align_multiline_strings(stmts, "\xaa")
    stmts = align_multiline_strings(stmts, "\xbb")

    # stmts.stmt = remove_strings(stmts.stmt, "\xaa")  # triple quote '''
    # stmts.stmt = remove_strings(stmts.stmt, "\xbb")  # triple quote """
    # stmts.stmt = split_fstrings(stmts.stmt, "'", '"')
    stmts.stmt = split_fstrings(stmts.stmt, '"', "'")
    stmts.stmt = remove_strings(stmts.stmt, "'")
    stmts.stmt = remove_strings(stmts.stmt, '"')

    stmts = clean_inline_comments(stmts, "#")
    stmts = clean_blanks(stmts)
    stmts = split_multi_statement_lines(stmts)
    stmts = align_context_block(stmts)
    stmts.stmt = replace_keywords(stmts.stmt)
    stmts.stmt = getback_combined_chars(stmts.stmt)
    stmts.stmt = enforce_4char_indents(stmts.stmt)

    return stmts



"""
Tucan module to handle the import parsing of all files.
"""
from typing import List
import json
from loguru import logger
from tucan.unformat_common import read_lines_with_encoding
from tucan.guess_language import guess_language
from tucan.imports_py import get_imports_py
from tucan.imports_ftn import get_imports_ftn
from tucan.imports_c import get_imports_ccpp
from tucan.unformat_py import unformat_py

from tucan.travel_in_package import rec_travel_through_package
from tucan.string_utils import (
    get_common_root_index_list,
    tokenize,
    best_front_match,
    best_inner_match,
)
from tucan.string_utils import generate_dict_paths


def imports_of_file(filename: str, verbose: bool = False) -> dict:
    """
    Extract the imports of a code file.

    Args:
        filename (str): Name of the file (with its path) to parse.

    Returns:
        dict: Imports functions stored by filename as key.
    """
    logger.info(f"Analyzing the imports for {filename}")

    code =  read_lines_with_encoding(filename)
    # not Lower cased !

    if filename.lower().endswith(".py"):
        logger.debug(f"Python code detected ...")
        imps = get_imports_py(unformat_py(code).to_code())
    elif filename.lower().endswith((".f", ".F", ".f77", ".F77", ".f90", ".F90")):
        logger.debug(f"Fortran code detected ...")
        imps = get_imports_ftn(code)
    elif filename.lower().endswith((".c", ".cpp", ".cc")):
        logger.debug(f"C/C++ code detected ...")
        imps = get_imports_ccpp(code)
    elif filename.lower().endswith((".h", ".hpp")):
        lang = guess_language(code)
        if lang in ["ccpp"]:
            logger.debug(f"C/C++ code detected ...")
            imps = get_imports_ccpp(code)
        elif lang in ["fortran"]:
            logger.debug(f"Fortran code detected ...")
            imps = get_imports_ftn(code)
        else:
            logger.debug(f"Language was not either C or Fortran, skipping...")
            imps = {}
    else:
        ext = filename.lower().split(".")[-1]
        logger.error(f"Extension {ext} not supported, exiting ...")
        imps = {}

    return imps, len(code)


def imports_of_repository(path: str) -> dict:
    paths = [path_.split("/") for path_ in rec_travel_through_package(path)]

    idx = get_common_root_index_list(paths)
    root = paths[0][:idx]
    # print(json.dumps(paths, indent=4))

    paths = [path_[idx:] for path_ in paths]

    def _full_path(root: list, path_: list) -> str:
        return "/".join(root) + "/" + "/".join(path_)

    tree_dict = {}
    for path_ in paths:
        current = tree_dict
        for bit in path_:
            if bit not in current:
                current[bit] = {}
            if bit.lower().endswith((".f", ".f77", ".f90")):
                fname = _full_path(root, path_)
                current[bit] = {key: {} for key in search_for_modules(fname)}
            current = current[bit]

    # tree_nob = Nob(tree_dict)
    tree_paths = generate_dict_paths(tree_dict)
    imports = {}
    sizes = {}
    for path_ in paths:
        local_imports = []
        imps_of_file, size_of_file = imports_of_file(_full_path(root, path_))
        for imps in imps_of_file.values():
            # logger.critical(path_)
            ref = find_import_ref(tree_paths, imps, orig_file="/".join(path_))
            if ref is not None:
                local_imports.append(ref)

        imports["/".join(path_)] = sorted(set(local_imports))
        sizes["/".join(path_)] = size_of_file

    return imports,sizes


def search_for_modules(filename: str) -> set:
    out = []
    code = [line.lower() for line in read_lines_with_encoding(filename)]
    for line in code:
        if line.lstrip().startswith("module"):
            out.append(tokenize(line)[1])
    return out


def remove_py_in_keys(dict_: dict) -> dict:
    """return the same dictionary but removong the .py in keys"""

    def _rec_no_py(indict_: dict) -> dict:
        """recursive, returning same dico w.o. .py in keys"""
        out = {}
        for key in indict_:
            out[key.replace(".py", "")] = _rec_no_py(indict_[key])
        return out

    return _rec_no_py(dict_)


def _find_import_by_file(tree_paths: list, file_ref: str) -> List[str]:
    hint = file_ref.split("/")

    # logger.warning(f"finding {hint} in {tree_paths}")
    imatch = best_inner_match(hint, tree_paths)
    # logger.warning(f"out is {imatch}")
    if imatch == []:
        return [f"__external__/{file_ref}"]

    out = ["/".join(stop_to_file(tree_paths[i])) for i in imatch]
    return out


def _find_import_for_ftn(tree_paths: list, mod_ref: str) -> List[str]:
    hint = mod_ref.split(".")
    # logger.info(f"In{hint}")
    # logger.info(f"TP{tree_paths}")
    # list_splited_paths = [_path.split("/") for _path in tree_paths]
    imatch = best_inner_match(hint, tree_paths)
    # logger.info(f"IM{imatch}")

    if imatch == []:
        return [f"__external__/{mod_ref}"]

    out = ["/".join(stop_to_file(tree_paths[i])) for i in imatch]
    return out


def _find_import_for_python(tree_paths: list, mod_ref: str) -> List[str]:
    hint = mod_ref.split(".")
    # logger.info(f"In{hint}")
    # logger.info(f"TP{tree_paths}")
    list_paths = ["/".join(path_) for path_ in tree_paths]
    # logger.info(f"LP{list_paths}")
    list_nopy_paths = [_path.replace(".py", "").split("/") for _path in list_paths]
    # logger.info(f"NP{list_nopy_paths}")

    imatch = best_inner_match(hint, list_nopy_paths)
    # logger.info(f"im{imatch}")

    if imatch == []:
        return [f"__external__/{mod_ref}"]

    out = ["/".join(stop_to_file(list_paths[i].split("/"))) for i in imatch]
    # logger.info(f"im{out}")
    return out


def find_import_ref(tree_paths: list, hint: str, orig_file: str = None) -> str:
    """Resolve import puzzles"""

    hint = hint.rstrip("*").rstrip(".")
    file_ref, mod_ref = hint.split("@")

    if file_ref == "__py__":
        matches = _find_import_for_python(tree_paths, mod_ref)
    elif file_ref == "__ftn__":
        matches = _find_import_for_ftn(tree_paths, mod_ref)
    else:
        matches = _find_import_by_file(tree_paths, file_ref)

    # logger.info(f"==> {hint} out {matches} in file {orig_file}")
    out = matches[0]
    if len(matches) > 1:
        if orig_file is None:
            out = matches[0]
        else:
            list_of = "\n".join(matches)
            logger.info(f"Front match {orig_file} in {list_of}")
            out = best_front_match(orig_file, matches)
            # logger.info(f"Found {out} from orig {orig_file.split('/')}")
    # logger.info(f"==> Found {out}")

    return out


def stop_to_file(path_l: List[str]) -> List[str]:
    extensions = [".py", ".f", ".f90", ".f77", ".c", ".cpp", ".cc", ".h", ".hpp"]
    out = []
    for item in path_l:
        out.append(item)
        if "." + item.lower().split(".")[-1] in extensions:
            break
    return out
